# AUTOGENERATED! DO NOT EDIT! File to edit: database.ipynb (unless otherwise specified).

__all__ = ['NeedUpdateIndex', 'Database', 'lambdaDumpToS3', 'lambdaSplitBranch', 'lambdaSaveS3', 'lambdaSingleQuery',
           'lambdaBranchUrl', 'lambdaAllQuery']

# Cell
from villaInvDatabase.helper import DatabaseHelper
from villaInvDatabase.s3 import DatabaseS3
from villaInvDatabase.query import Querier
from villaInvDatabase.update import Updater
import pandas as pd
from datetime import datetime
from pynamodb.models import Model
from pynamodb.attributes import UnicodeAttribute, NumberAttribute, JSONAttribute, BooleanAttribute, BinaryAttribute
from pynamodb.indexes import GlobalSecondaryIndex, AllProjection
from awsSchema.apigateway import Response, Event
from botocore.config import Config
from s3bz.s3bz import S3
from linesdk.slack import SlackBot
from pprint import pprint

import pickle, json, boto3, bz2, requests, validators, os, logging

# Cell
try:
  SLACK = os.environ.get('SLACK')
  DATABASE_TABLE_NAME = os.environ['DATABASE_TABLE_NAME']
  INVENTORY_BUCKET_NAME = os.environ['INVENTORY_BUCKET_NAME']
  INPUT_BUCKET_NAME = os.environ['INPUT_BUCKET_NAME']
  REGION = os.environ['REGION']
  ACCESS_KEY_ID = None
  SECRET_ACCESS_KEY = None
except Exception as e:
  print(f'error, missing environment variables \n{e}')
  DATABASE_TABLE_NAME = None
  INVENTORY_BUCKET_NAME = None
  INPUT_BUCKET_NAME = None
  ACCESS_KEY_ID = None
  SECRET_ACCESS_KEY = None
  REGION = 'ap-southeast-1'
try:
  DAX_ENDPOINT = os.environ['DAX_ENDPOINT']
except:
  DAX_ENDPOINT = None
  print('dax endpoint missing')


# Cell
class NeedUpdateIndex(GlobalSecondaryIndex):
  class Meta:
    index_name = 'need-update'
    projection = AllProjection()
    read_capacity_units = 1
    write_capacity_units = 1
  needUpdate = UnicodeAttribute(default=False, hash_key=True)

# Cell
# dont forget to import dependent classes from the relevant notebooks
class Database(Model, DatabaseS3, Updater, Querier, DatabaseHelper):
  class Meta:
    aws_access_key_id = ACCESS_KEY_ID
    aws_secret_access_key = SECRET_ACCESS_KEY
    table_name = DATABASE_TABLE_NAME
    region = REGION
    billing_mode='PAY_PER_REQUEST'
    dax_read_endpoints = [DAX_ENDPOINT] if DAX_ENDPOINT else None
    dax_write_endpoints = [DAX_ENDPOINT] if DAX_ENDPOINT else None

  ib_prcode = UnicodeAttribute(hash_key=True, default = '')
  lastUpdate = NumberAttribute( default = 0)
  inventory = JSONAttribute(default={})
  needUpdate = UnicodeAttribute(default= 'N')
  needUpdateIndex = NeedUpdateIndex()
  TRUE = 'Y'
  FALSE = 'N'

  def __repr__(self):
    invDict = self.inventory
    invDict['ib_prcode'] = self.ib_prcode
    return json.dumps(invDict)
  def setNoUpdate(self, batch = None):
    self.needUpdate = self.FALSE
    if batch:
      return batch.save(self)
    else:
      return self.save()
  def setUpdate(self):
    self.needUpdate = self.TRUE
    return self.save()

# Cell
def lambdaDumpToS3(event, _):
  result = Database.dumpToS3()
  try:
    bot = SlackBot(SLACK)
    bot.sendMessage(f'{result}')
  except:
    logging.exception('unable to send slack')
  return Response.getReturn(body={'savingResult': result})

# Cell
def lambdaSplitBranch(event, _):
  result = Database.splitBranches(bucket = INVENTORY_BUCKET_NAME)
  try:
    bot = SlackBot(SLACK)
    bot.sendMessage(f'{result}')
  except:
    logging.exception('unable to send slack')
  return Response.getReturn(body={'savingResult': result})

# Cell
def lambdaSaveS3(events, _):
  key = events.get('key')
  updateResult = Database.updateS3Input(
    inputBucketName=INPUT_BUCKET_NAME, key= key)
  return Response.getReturn(body = {'saveResult': updateResult})

# Cell
def lambdaSingleQuery(event, _):
  body = Event.from_dict(event).getBody()
  result = Database.singleProductQuery(body).inventory
  return Response.getReturn(body = result)


# Cell
def lambdaBranchUrl (events, _):
  if 'body' not in events.keys(): return Response.getReturn(statusCode=400, body = {'error': 'missing body tag'})
  branch = Event.from_dict(events).getBody()['branch']
  branchURL = Database.branchQuery('1000', bucket = INVENTORY_BUCKET_NAME)
  print(branchURL[:20])
  return Response.getReturn(body = {'url': branchURL})

# Cell
def lambdaAllQuery (events, _):
  url = Database.allQuery(bucket = INVENTORY_BUCKET_NAME, user=USER, pw=PW)
  print(url[:20])
  return Response.getReturn(body = {'url': url})



# AUTOGENERATED! DO NOT EDIT! File to edit: update.ipynb (unless otherwise specified).

__all__ = ['Updater']

# Cell
import pandas as pd
from datetime import datetime
import pickle, json , boto3, zlib, os, logging, traceback

# Cell
try:
#   INVENTORY_BUCKET_NAME = os.environ['INVENTORY_BUCKET_NAME']
  INPUT_BUCKET_NAME = os.environ['INPUT_BUCKET_NAME']
except Exception as e:
  print(f'missing environment variable {e} in update NB')
#   INVENTORY_BUCKET_NAME = None
  INPUT_BUCKET_NAME = None

# Cell
class Updater:
  @classmethod
  def valueUpdate(cls, input):
    '''
    update products in the database, first use the helper.groupbyprcode to treat the data
    input structure is the foloowing
    - iprcode:
        ibrcode:
          ib_cf_qty: Int
          lastUpdate: Float
          new_ib_bs_stock_cv: Int
    '''
    itemsUpdated = {'success':0, 'failure': 0, 'failureMessage':[]}

    logging.info(f'there are {len(list(input.keys()))} products to update')

    # loop through each product
    for prcode, value in input.items():

      incumbentBr = None
      # check if product is in the database, if not, create an empty class with the product code
      incumbentBr = next(cls.query(prcode),cls(ib_prcode = prcode, inventory = {}))
      logging.info(f'incumbentBr is {incumbentBr}\n, prcode is {prcode}')

      # drop the data with no update
      filteredBr = {
          brcode : v2
          for brcode, v2 in value.items()
          if v2['ib_cf_qty'] != incumbentBr.inventory.get(brcode ,{}).get('ib_cf_qty',None)
      }
      logging.info(f'filteredBr is {filteredBr}')


      # check if there are data to update
      if filteredBr:
        # put the latest update into the lastupdate index
        incumbentBr.lastUpdate = max( v.get('lastUpdate') for v in filteredBr.values() )
        incumbentBr.inventory = (lambda d: d.update(filteredBr) or d)(incumbentBr.inventory)
        incumbentBr.inventory['lastUpdate'] = incumbentBr.lastUpdate
        incumbentBr.inventory['ib_prcode'] = incumbentBr.ib_prcode
        incumbentBr.needUpdate = cls.TRUE

        logging.info(f'saving result {incumbentBr}')
        # try to save the result
        saveResult = incumbentBr.save()
        # saveResult = batch.save(incumbentBr)

        # record the saving success
        if saveResult.get('ConsumedCapacity'):
          itemsUpdated['success'] += 1
        else:
          itemsUpdated['failure'] += 1
          itemsUpdated['failureMessage'].append(saveResult)

    return itemsUpdated

  @classmethod
  def bulkUpdate(cls,input, **kwargs):
    '''
    update products in the database, first use the helper.groupbyprcode to treat the data
    input structure is the foloowing
    - iprcode:
        ibrcode:
          ib_cf_qty: Int
          lastUpdate: Float
          new_ib_bs_stock_cv: Int
    '''
    itemsUpdated = {'success':0, 'failure': 0, 'failureMessage':[]}

    logging.info(f'there are {len(list(input.keys()))} products to update')

    db = cls.loadFromS3(**kwargs)

    with cls.batch_write() as batch:
      # loop through each product
      for prcode, value in input.items():

        incumbentBr = None
        # check if product is in the database, if not, create an empty class with the product code
  #       incumbentBr = next(cls.query(prcode),cls(ib_prcode = prcode, inventory = {}))
        # do this with s3 database instead of dynamodb to speed up
        incumbentBr = cls(inventory = db.get(prcode) or {}, ib_prcode = prcode)
        logging.info(f'incumbentBr is {incumbentBr}\n, prcode is {prcode}')

        # drop the data with no update
        filteredBr = {
            brcode : v2
            for brcode, v2 in value.items()
            if v2['ib_cf_qty'] != incumbentBr.inventory.get(brcode ,{}).get('ib_cf_qty',None)
        }
        logging.info(f'filteredBr is {filteredBr}')


        # check if there are data to update
        if filteredBr:
          # put the latest update into the lastupdate index
          incumbentBr.lastUpdate = max( v.get('lastUpdate') for v in filteredBr.values() )
          incumbentBr.inventory = (lambda d: d.update(filteredBr) or d)(incumbentBr.inventory)
          incumbentBr.inventory['lastUpdate'] = incumbentBr.lastUpdate
          incumbentBr.inventory['ib_prcode'] = incumbentBr.ib_prcode
          incumbentBr.needUpdate = cls.TRUE

          logging.info(f'saving result {incumbentBr}')
          # try to save the result
#           saveResult = incumbentBr.save()
          try:
            saveResult = batch.save(incumbentBr)
            itemsUpdated['success'] += 1
          except:
            itemsUpdated['failure'] += 1
            itemsUpdated['failureMessage'].append(traceback.format_exc())
    return itemsUpdated



  @classmethod
  def updateLambdaInput(cls, input):
    '''
    update products in the database by first grouping the data from lambda
    input
    - ib_prcode: String
      ib_brcode: String
      ib_cf_qty: Int
      new_ib_bs_stock: int
    '''
    groupedInput = cls.Helper.groupByProduct(input)
    return cls.valueUpdate(groupedInput)

  @classmethod
  def updateS3Input(cls, inputBucketName = INPUT_BUCKET_NAME, key = '', **kwargs):
    s3Result = cls.loadFromS3(bucketName= inputBucketName, key = key, **kwargs)
    transformedS3Result = [{
        'ib_prcode': item.get('ib_prcode') or item.get('prcode'),
        'ib_brcode': item.get('ib_brcode') or item.get('brcode'),
        'ib_cf_qty': item.get('ib_cf_qty'),
        'new_ib_vs_stock_cv': item.get('new_ib_vs_stock_cv')
        } for item in s3Result]

    logging.info(f' s3 result is {transformedS3Result}')
    groupedInput = cls.Helper.groupByProduct(transformedS3Result)
    updateResult = cls.bulkUpdate(groupedInput, **kwargs)
    return updateResult



# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"DEFAULTCOLS": "database2.ipynb",
         "PandasDataFrameAttribute": "database2.ipynb",
         "Database": "database2.ipynb",
         "updateWithDfs": "database2.ipynb",
         "lambdaPresignUpload": "database2.ipynb",
         "lambdaIngestUpload": "database2.ipynb",
         "lambdaSingleBranchQuery": "database2.ipynb",
         "lambdaAllBranchesQuery": "database2.ipynb",
         "DatabaseHelper": "helper.ipynb",
         "Querier": "query.ipynb",
         "DatabaseS3": "s3.ipynb",
         "Updater": "update.ipynb"}

modules = ["database2.py",
           "helper.py",
           "query.py",
           "s3.py",
           "update.py"]

doc_url = "https://thanakijwanavit.github.io/villaPriceDatabase/"

git_url = "https://github.com/thanakijwanavit/villaPriceDatabase/tree/master/"

def custom_doc_links(name): return None



# AUTOGENERATED! DO NOT EDIT! File to edit: query.ipynb (unless otherwise specified).

__all__ = ['Querier']

# Cell
import pandas as pd
from datetime import datetime
from s3bz.s3bz import S3
import pickle, json , boto3, zlib, os

# Cell
try:
  INVENTORY_BUCKET_NAME = os.environ['INVENTORY_BUCKET_NAME']
except Exception as e:
  print(f'missing environment variable {e} in query NB')
  INVENTORY_BUCKET_NAME = None

# Cell
class Querier:
  @classmethod
  def singleProductQuery(cls, input):
    if not cls.validateInputQuery(['ib_prcode'] , input): return f"error input {input}"
    return next(cls.query(input.get('ib_prcode')),{})

  @classmethod
  def branchQuery(cls, branchId:str, bucket = INVENTORY_BUCKET_NAME, **kwargs):
    key = branchId
    result = S3.presign(key, bucket = bucket, **kwargs)
    return result
  @classmethod
  def allQuery(cls, key = 'allData', bucket = INVENTORY_BUCKET_NAME, **kwargs):
    result = S3.presign(key, bucket = bucket, **kwargs)
    return result

  @staticmethod
  def validateInputQuery(keys: list, input:dict):
    '''
      check if input query contains the valid key
      data should have the following structure
      key is a list of keys to check

      ib_prcode: String?
      ib_brcode: String?

      option, one of or both of the ib_procde must be present
    '''
    for key in keys:
      if key not in input.keys():
        raise ValueError(f"key {key} is missing from the input")
      if not input.get(key).isdigit():
        raise ValueError(f'key is not convertable to in {input.get(key)}')
    return True


# AUTOGENERATED! DO NOT EDIT! File to edit: s3.ipynb (unless otherwise specified).

__all__ = ['DatabaseS3']

# Cell
import os, logging
from s3bz.s3bz import S3

# Cell
try:
  INVENTORY_BUCKET_NAME = os.environ['INVENTORY_BUCKET_NAME']
except Exception as e:
  print(f'missing environment variable {e} in Database s3 NB')
  INVENTORY_BUCKET_NAME = None

# Cell
class DatabaseS3:
  @classmethod
  def loadFromS3(cls, bucketName= INVENTORY_BUCKET_NAME, key = 'allData', **kwargs):
    logging.info(f'loading from {bucketName}')
    logging.info(f'user is {kwargs.get("user")}')
    return S3.load(key=key, bucket = bucketName,  **kwargs)

  @classmethod
  def dumpToS3(cls, bucketName= INVENTORY_BUCKET_NAME, key = 'allData', limit= 1000, rebase = False, **kwargs):
    ''' upload changes to s3'''
    if rebase:
      changeList = list(cls.needUpdateIndex.scan())
    else:
      changeList = list(cls.needUpdateIndex.query(cls.TRUE, limit=limit))
    # check for changes otherwise, skip
    if not changeList:
      return {
        "newDataSaved": False,
        "numberOfProducts": 0,
        "message": f'no changes to database'
      }

    # get original data
    allData = cls.loadFromS3(bucketName = bucketName, key = key, **kwargs)
    logging.debug(f'all data is {len(allData)}')
    logging.debug(f'{len(changeList)} changes to update')
    with cls.batch_write() as batch:
      changes = {
                  item.ib_prcode:(
                    item.inventory,
                    item.setNoUpdate(batch=batch)
                  )[0]
                  for item in changeList
                }
    logging.debug(f'change filtration has completed')
    logging.debug(f'all data is {len(allData)}\n changes are {len(changes)}')
    if changes:
      logging.debug(f'updating')
      allData.update(changes)
      logging.debug(S3.save(key = 'allData',
                  objectToSave = allData,
                  bucket = bucketName, **kwargs)
      )
    else:
      logging.debug('no changes to update')

    numberOfChanges = len(list(changes.items()))
    numberOfProducts = len(list(allData.keys()))
    if numberOfChanges == limit: logging.warning('more changes to load')
    return {
      "newDataSaved": numberOfChanges != 0,
      "numberOfProducts": numberOfProducts,
      "message": f"saved {numberOfProducts} products,\
            updated {numberOfChanges}"
    }



# AUTOGENERATED! DO NOT EDIT! File to edit: database2.ipynb (unless otherwise specified).

__all__ = ['DEFAULTCOLS', 'PandasDataFrameAttribute', 'Database', 'updateWithDfs', 'lambdaPresignUpload',
           'lambdaIngestUpload', 'lambdaSingleBranchQuery', 'lambdaAllBranchesQuery']

# Cell
from .helper import DatabaseHelper
from .s3 import DatabaseS3
from .query import Querier
from .update import Updater
import pandas as pd
from datetime import datetime
from pynamodb.models import Model
from pynamodb.attributes import UnicodeAttribute, NumberAttribute, JSONAttribute, BooleanAttribute, BinaryAttribute, Attribute
from pynamodb.indexes import GlobalSecondaryIndex, AllProjection
from pynamodb.constants import BINARY
from awsSchema.apigateway import Response, Event
from botocore.config import Config
from s3bz.s3bz import S3
from linesdk.slack import SlackBot
from pprint import pprint
from nicHelper.wrappers import add_class_method, add_method
from nicHelper import pdUtils
from nicHelper.pdUtils import getDfHash
from nicHelper.dictUtil import hashDict
from io import BytesIO

import pickle, json, boto3, bz2, requests, validators, os, logging, sys, zlib

# Cell

DEFAULTCOLS = json.loads(os.environ.get('DEFAULTCOLS') or '[]')
try:
  SLACK = os.environ.get('SLACK')
  DATABASE_TABLE_NAME = os.environ['DATABASE_TABLE_NAME']
  PRICE_BUCKET_NAME = os.environ['PRICE_BUCKET_NAME']
  INPUT_BUCKET_NAME = os.environ['INPUT_BUCKET_NAME']
  REGION = os.environ['REGION']
  ACCESS_KEY_ID = None
  SECRET_ACCESS_KEY = None
except Exception as e:
  print(f'error, missing environment variables \n{e}')
  DATABASE_TABLE_NAME = None
  INVENTORY_BUCKET_NAME = None
  INPUT_BUCKET_NAME = None
  ACCESS_KEY_ID = None
  SECRET_ACCESS_KEY = None
  REGION = 'ap-southeast-1'
try:
  DAX_ENDPOINT = os.environ['DAX_ENDPOINT']
except:
  DAX_ENDPOINT = None
  print('dax endpoint missing')


# Cell
# dont forget to import dependent classes from the relevant notebooks
class PandasDataFrameAttribute(Attribute):
  attr_type = BINARY
  def serialize(self, value: pd.DataFrame)->bin:
    bio = BytesIO()
    floatVal:pd.DataFrame = value.astype(float)
    floatVal[DEFAULTCOLS].to_feather(bio)
    data:bin = bio.getvalue()
    compressedData:bin = zlib.compress(data)
    return compressedData
  def deserialize(self, compressedData: bin)->pd.DataFrame:
    data = zlib.decompress(compressedData)
    bio = BytesIO(data)
    df: pd.DataFrame = pd.read_feather(bio)[DEFAULTCOLS]
    return df
class Database(Model):
  class Meta:
    table_name = DATABASE_TABLE_NAME
    region = REGION
    billing_mode='PAY_PER_REQUEST'
    dax_read_endpoints = [DAX_ENDPOINT] if DAX_ENDPOINT else None
    dax_write_endpoints = [DAX_ENDPOINT] if DAX_ENDPOINT else None

  brcode = UnicodeAttribute(hash_key=True, default = 0)
  lastUpdate = NumberAttribute(default=datetime.now().timestamp())
  data = PandasDataFrameAttribute()

  def __repr__(self):
    return self.data.head().to_string()
  def size(self):
    return f'data is size {sys.getsizeof(self.data)/1e3} kB'





# Cell
@add_class_method(Database)
def updateWithDfs(cls, df: pd.DataFrame, defaultDf= pd.DataFrame({i:[] for i in DEFAULTCOLS})):
  results = []
  ##### split df based on brcode
  fdf = {}
  brcodes = df['brcode'].unique()
  for brcode in brcodes:
    mask = df['brcode'] == brcode
    fdf[brcode] = df[mask]

  ##### loop through each df and update in the database
  with cls.batch_write() as batch:
    for brcode, data in fdf.items():
      db:cls = next(cls.query(str(brcode)),None) or cls(brcode=str(brcode), data=defaultDf)
      db.lastUpdate = datetime.now().timestamp()
      df0 = db.data.set_index('cprcode')
      df1 = data.set_index('cprcode')
      db.data = df0.combine_first(df1).reset_index()
      print(db.size())
      result = batch.save(db)
  return True

# Cell
def lambdaPresignUpload(event, *args):
  key:str = Event.parseBody(event)['key']
  presigned:dict = S3.presignUpload(bucket=PRICE_BUCKET_NAME, key = key, expiry = 1000)
  return Response.returnSuccess(body = presigned)

# Cell
def lambdaIngestUpload(event, *args):
  body = Event.parseBody(event)
  key = body['key']
  dtype = body.get('dtype') or 'json'
  path = '/tmp/input'
  S3.loadFile(key = key, path = path, bucket = PRICE_BUCKET_NAME)
  if dtype == 'json':
    df = pd.read_json(path, dtype=str)
    print(f'reading jsoon {df.head()}')
  if dtype == 'feather':
    df = pd.read_feather(path)
    print(f'reading feather{df.head()}')
  result = Database.updateWithDfs(df=df)
  return Response.returnSuccess(df.head().to_dict())

# Cell
def lambdaSingleBranchQuery(event:dict, *args):
  def queryDatabase(brcode:str)->pd.DataFrame:
    df:pd.DataFrame = next(Database.query(brcode), Database(data=pd.DataFrame(dtype=int))).data
    return df

  def filterDb(cprcodes:List[int], df:pd.DataFrame):
    '''filter database based on the cprcodes'''
    filteredDf:pd.DataFrame = df[df['cprcode'].isin(cprcodes)]
    return filteredDf

  def saveDbToFile(df:pd.DataFrame, path:str, format_:str):
    if format_ == 'feather': df.to_feather(path=path)
    else: dbDict:dict = df.to_json(path,  orient='split')

  def saveFileToS3(key:str, path:str, bucket:str=PRICE_BUCKET_NAME):
    S3.saveFile(key=key,path=path,bucket=PRICE_BUCKET_NAME)
  ##### setup
  path:str = '/tmp/tempFile'
  body:dict = Event.parseBody(event)
  key:str = hashDict(body)
  brcode:str = str(body['brcode'])
  cprcodes:List[int] = [int(i) for i in body.get('cprcodes')]
  format_:str = body.get('format') or 'json'

  ###### main #####
  df = queryDatabase(brcode) # get data from db
  if cprcodes:
    df = filterDb(cprcodes, df) # filter the data
  saveDbToFile(df, path, format_) # save the df to file
  saveFileToS3(key, path) # move local file to s3
  url = S3.presign(key=key,path=path,bucket=PRICE_BUCKET_NAME) #presign the saved file

  return Response.returnSuccess({'url': url}) # retur the url

# Cell
from nicHelper.dictUtil import hashDict
def lambdaAllBranchesQuery(event, *args):
  def getDb():
    result = Database.scan()
    dfs = [i.data for i in result]
    return dfs
  def mergeDf(dfs:List[pd.DataFrame])->pd.DataFrame:
    df:pd.DataFrame = pd.concat(dfs).astype(int).reset_index(drop=True)
    df = df.drop_duplicates(subset=['cprcode', 'brcode'],keep='last')
    return df
  def filterDf(df:pd.DataFrame, cprcodes:List[int]):
    fdf:pd.DataFrame = df[df['cprcode'].isin(cprcodes)]
    return fdf


  def dfToUrl(df:pd.DataFrame, path:str, key:str, format_:str, bucket:str):
    '''save df to s3 then create a presigned link'''
    def saveDbToFile(df:pd.DataFrame, path:str, format_:str):
      if format_ == 'feather':
        df.reset_index(drop=True).to_feather(path=path)
      else: dbDict:dict = df.to_json(path,  orient='split')

    def saveFileToS3(key:str, path:str, bucket:str=PRICE_BUCKET_NAME):
      S3.saveFile(key=key,path=path,bucket=PRICE_BUCKET_NAME)

    ### main ###
    saveDbToFile(df, path, format_)
    saveFileToS3(key, path, bucket=bucket)
    url = S3.presign(key=key,path=path,bucket=bucket)
    return url

  #### setup
  body:dict = Event.from_dict(event).getBody()
  if 'cprcodes' in body:
    cprcodes:List[int] = [int(i) for i in body.get('cprcodes')]
  else:
    cprcodes = []
  outputFormat:str = body.get('format') or 'json'
  key:str = hashDict(body)
  path = '/tmp/tmpFile'

  #### main ######
  dfs:List[pd.DataFrame] = getDb()
  df:pd.DataFrame = mergeDf(dfs)
  if cprcodes: # filter df is cprcode exists
    df = filterDf(df, cprcodes)
  url = dfToUrl(df=df, path=path, key=key, format_=outputFormat, bucket=PRICE_BUCKET_NAME) # push df to url

  return Response.returnSuccess(body = {'url':url})



# AUTOGENERATED! DO NOT EDIT! File to edit: helper.ipynb (unless otherwise specified).

__all__ = ['DatabaseHelper']

# Cell
import pandas as pd
from datetime import datetime
import pickle, json , boto3, zlib, os, logging
from s3bz.s3bz import S3

# Cell
class DatabaseHelper:
  class Helper:
    @staticmethod
    def timer(t0):
      return (datetime.now()-t0).total_seconds()
    @staticmethod
    def groupByProduct(data):
      df = pd.DataFrame(data)
      gprcode = df.groupby('ib_prcode')
      # group by ib_prcode, then ib_brcode, add timestamp as well
      res = gprcode.agg(list).apply(lambda x: {
            k:v for k,v in zip(
                x['ib_brcode'], [
                  {
                    'ib_cf_qty': int(qty),
                    'new_ib_bs_stock_cv': int(new),
                    'lastUpdate' : datetime.now().timestamp()
                  }
                  for qty,new in zip(x['ib_cf_qty'], x['new_ib_vs_stock_cv'])
                ]
              )
          } , axis = 1)
      return res.to_dict()
    @staticmethod
    def setUpdateAll():
      return list(item.setUpdate() for item in Database.scan())

    @staticmethod
    def iteratorToInvDict(data):
      return {item.ib_prcode:item.inventory for item in data}

  @classmethod
  def splitBranches(cls, key='allData',bucket = '', **kwargs):
    '''
    split database items from s3 into individual branch
    '''
    t0 = datetime.now() # record time
    if not bucket:
      logging.warning('bucket name is empty')
      return 'empty bucket name'

    data = S3.load(key=key , bucket = bucket, **kwargs)# load allData
    logging.info(f'took {cls.Helper.timer(t0)} to load input s3 object ')
    if not data: return 'items are empty' # check if object is blank

    # create cleanedSplitDict

    df = pd.DataFrame(data).drop(['lastUpdate', 'ib_prcode'])
    splitDict = df.T.fillna('')
    logging.info(f'took {cls.Helper.timer(t0)} to transpose and fillna')
    splitDict = splitDict.to_dict()
    logging.info(f'took {cls.Helper.timer(t0)} to convert df to dict')
    cleanedSplitDict = {brcode:
      { k:v for k,v in inv.items()  if v}
      for brcode, inv in splitDict.items()}
    logging.info(f'took {cls.Helper.timer(t0)} to dict comprehension')
    logging.info(f'took {cls.Helper.timer(t0)} to load and splitDict')

    logging.info(f'first row is {next(iter(cleanedSplitDict))}')
    saveResults = {'success':0, 'failure':0,'errorMessage':[]}
    for k,v in cleanedSplitDict.items():
      logging.debug(k,v)
      saveResult = S3.save(key = k,
                           objectToSave = v,
                           bucket = bucket,
                           **kwargs)
      logging.info(f'{cls.Helper.timer(t0)} s3 branch {k} saved')
      if saveResult: saveResults['success'] += 1
      else: saveResults['failure'] += 1
    return saveResults